%%           -*-Mode: prolog;-*-

%%:- ensure_loaded(eacl99).      %% without rx
%%:- ensure_loaded(replace).     %% Mohri-Sproat style
:- ensure_loaded(replace_nm).    %% no special treatment of markers

:- multifile macro/2.
:- multifile rx/2.

%% set spy-points on certain operators
:- user:bb_put(fsa_rx_spy:parse,on).
:- user:bb_put(fsa_rx_spy:l,on).
%% our atomic symbols are `words' separated by `spaces':
:- user:bb_put(fsa:symbol_separator,32).
:- user:bb_put(fsa_regex_cache:l,on).
%% generate strings without considerations to the length. 
:- user:bb_put(fsa:length_max,0).

%% USAGE:
%% fsa -tk -aux syntax debug=1 
%% regex: parse(ExNo)
%%
%% first time: very slow.
%% After that, all the level/1 operators are cached. Therefore, further
%% examples are much faster; about 3 seconds.
%% The implication is that it is much cheaper to compute the compositions
%% on line (for a particular example) then to pre-compute the compositions
%% for the general case. But still not cheap enough.

%% the input is assumed to be tagged already; cf. zin? for examples.
%% every word X is replaced by the four symbols: Tag [ X ] 
%% where Tag is the corresponding tag, and X is the word itself.
%% de man met de vrouw
%% det [ de ] noun [ man ] prep [ met ] det [ de ] noun [ vrouw ]

%% this is terrible:
macro(parse, l(m) o l(0) o l(1) o l(2) o l(3) o l(4) o l(5) ).

macro(parse(Int), show_tree(parse0(Int))).

macro(parse0(I),
      minimize(range(zin(I) o l(m) o l(0) o l(1) o l(2) o
	                      l(3) o l(4) o l(5)     ))).

macro(sparse(Int), show_tree(sparse0(Int))).

macro(sparse0(I),
      minimize(range(zin(I) o wiaparser   ))).

macro(wiaparser,
%             l(m)
%               o
  replace(( adj --> [d(adv), d(adj)]))
               o
  replace((  np --> [d(art)^,d(num)^,d(adj)*,d(n)+]))
               o
  replace((  pp --> [d(prep),d(np)]))
               o
  replace((  np --> [d(np),d(pp)+]))
     ).

%% multi-word-units
macro(l(m),replace({ ( prep --> [dw(prep,{'Ten',ten}),
				 dw(n,opzichte),
				 dw(prep,van)]                     ),
		     ( prep --> [dw(prep,{'In',in}),
				 dw(n,verband),
				 dw(prep,met)]                     ),
		     ( prep --> [dw(prep,{'In',in}),
				 dw(n,plaats),
				 dw(prep,van)]                     )
		   })).

macro(mmm,  (prep --> [dw(prep,in),
		       dw(n,verband),
		       dw(prep,met)])
     ).

macro(l(0),replace({ (  adj --> [d(adv),d(adj)]                    ),
		     (    v --> [dw(prep,te),d(v)]                 )
		   })).

macro(l(1),replace({ (  adj --> [d(adj),dw(conj,{en,of}),d(adj)]   )
		   })).

macro(l(2),replace({ (   np --> [d(art)^,d(num)^,d(adj)*,d(n)+]    )
		   })).

macro(l(3),replace({ (   np --> [d(np),dw(conj,{en,of}),d(np)]     )
		   })).

macro(l(4),replace({ (   pp --> [d(prep),d(np)]                    )
                   })).

macro(l(44),replace({ (   np --> [d(np),d(pp)]                    )
                   })).

macro(l(5),replace({ (   np --> [d(np),d(pp)+]                     )
		   })).

%% A --> Ds: introduce A brackets around Ds. 
macro((A --> Ds), [[] x [A,'['], Ds, []:']']).

%% d(Cat) daughter category: Cat, top-leftbracket, anything, top-rightbracket
%% where anything does not contain top-rightbracket. 
%% The top-brackets are altered to embedded bracket. 
macro(d(Cat), [Cat, '[':'(', (? - ']')*,']':')'] ).


%% dw(Cat,Sub) matches Cat, top-leftbracket, Sub, top-rightbracket
%% The top-brackets are altered to embedded bracket.
%% dw/2 can be used instead of d/1 if you want to specify a pattern for
%% the embedded structure of Cat. In that case, you have to make sure your
%% self that this doesn't contain any top-level brackets.
macro(dw(Cat,Word), [Cat, '[':'(',Word,']':')']).

%% dw(Cat) matches Cat, top-leftbracket, single-word, top-rightbracket
macro(dw(Cat),dw(Cat,?)).

%% STUFF to display output as a real tree!
%% show_tree/1 is only interesting for its side-effect!
rx(show_tree(X),F) :-
    fsa_regex:rx(X,F),
    fsa_interpreter:fsa_accepts(List,F),
    list_to_tree(List,Tree),
    display_tree(Tree).

list_to_tree(List,tree(zin,_,Ds)):-
    list_to_tree_ds(Ds,List,[]).

list_to_tree_ds([H|T]) -->
    list_to_tree_el(H),
    list_to_tree_ds0(T).

list_to_tree_ds0([H|T]) -->
    list_to_tree_el(H),
    list_to_tree_ds0(T).
list_to_tree_ds0([]) --> [].

list_to_tree_el(tree(Label,_,Ds)) -->
    [Label],
    ['['],
    list_to_tree_ds(Ds),
    [']'].
list_to_tree_el(tree(Label,_,Ds)) -->
    [Label],
    ['('],
    list_to_tree_ds(Ds),
    [')'].
list_to_tree_el(tree(Label,_,[])) -->
    [Label],
    {  \+ bracket(Label) }.

bracket('[').
bracket(']').
bracket('(').
bracket(')').

:- (   user:bb_get(user:display_tree,_)
   ->  true
   ;   user:bb_put(user:display_tree,default)
   ).

display_tree(Tree) :-
    user:bb_get(user:display_tree,Flag),
    display_tree(Flag,Tree).

%% using CLiG for Tcl/Tk trees
display_tree(hdrug_clig,Tree) :-
    use_module('/opt_local/opt/prolog/Hdrug/Prolog/hdrug_clig'),
    (	user:current_predicate(hdrug_library,hdrug_library(_))
    ->	true
    ;   assertz(user:hdrug_library('/opt_local/opt/prolog/Hdrug/Tcl'))
    ),
    clig_tree(Tree).

display_tree(default,Tree) :-
    pp_tree(Tree,0).

%%%%%%%%%%% EXAMPLES %%%%%%%%%%%%%%%%%
%%%%%%%% Eindhoven Corpus %%%%%%%%%%%%
%%%%%%%% simplified tags %%%%%%%%%%%%%

:- ensure_loaded(tagged_examples).


pp_tree(tree(Label,_,Ds),Tab0):-
    tab(user_error,Tab0),
    format(user_error,"~w~n",[Label]),
    Tab is +(Tab0,3),
    pp_tree_ds(Ds,Tab).

pp_tree_ds([],_).
pp_tree_ds([H|T],Tab):-
    pp_tree(H,Tab),
    pp_tree_ds(T,Tab).

