:- use_module('/net/aps/64/lib/fsa/sic/fsa_library').

:- use_module(uniform).

input_to_key_string(Input,Key,String) :-
    lists:append(Key,[124|String],Input).

output_all([],_,Nr,Nr).
output_all([StringOut|T],Key,Nr0,Nr) :-
    fsa_interpreter:list_of_symbols_to_string(StringOut,Result,in),
    format("~s#~w|~s~n",[Key,Nr0,Result]),
    Nr1 is Nr0 + 1,
    output_all(T,Key,Nr1,Nr).

main :-
    format(user_error,"reading transducer...~n",[]),
    fsa_read_file(fast,'confusable_tttt.t',Transducer),
    format(user_error,"reading transducer ok~n",[]),
    repeat,
    read_line(Input),
    (   Input == end_of_file
    ->  true
    ;   input_to_key_string(Input,Key,String),
	length(String,Len),
	atom_codes(Atom,String),
	format(user_error,"compiling regex ..~n",[]),
	fsa_regex_compile(determinize(range(compose(word(Atom),fa(Transducer)))),Fa),
	fsa_uniform(Fa,FaU),
	format(user_error,"regex ok~n",[]),

	findall(StringOut,
		fsa_sample(FaU,Len,StringOut), % length of input = number of variants
		List0
	       ),

	sort(List0,List),
	output_all(List,Key,1,_Nr),
	format("~s#0|~s~n",[Key,String]),
	fail
    ),
    !.  % repeat
	
fsa_sample(Fa,Max,Symbols) :-
    fsa_data:fsa_symbols_decl(Fa,r(Module,_)),
    fsa_data:index_transitions(Fa,Table),
    user:bb_put(fsa_interpreter:nrsol,0),
    repeat,
    fsa_interpreter:fsa_sample(Fa,Table,List,recognizer),
    fsa_interpreter:random_symbols_from_preds(List,Symbols,Module),
    fsa_interpreter:inc(nrsol),
    user:bb_get(fsa_interpreter:nrsol,Sols),
    (   Sols >= Max
    ->  !  % repeat
    ;   true
    ).

:- main.

