:- use_module('/net/aps/64/lib/fsa/sic/fsa_library').

%% uniform(Fa0,Fa)
%%
%% for a deterministic recognizer Fa0 defining a *finite* language
%% (i.e. acyclic deterministic automaton)
%% we construct a weighted recognizer, in such a way
%% that every string is assigned the same weight
%% (probability) by that weighter recognizer.
%%
%% procedure: for every state, compute the
%% number of strings of the right language of
%% that state.
%%
%%

test(Fa) :-
    fsa_regex:fsa_regex_compile({a,[a,b],[a,c],[a,b,a],[a,b,b],[a,b,c],[a,b,c,d]},Fa0),
    fsa_uniform(Fa0,Fa).

test1(Fa) :-
    fsa_regex:fsa_regex_compile([a,{b,c},d],Fa0),
    fsa_uniform(Fa0,Fa).

fsa_uniform(Fa0,Fa) :-
    retractall(count_target_done(_,_)),
    fsa_data:fsa_type(Fa0,recognizer),
    fsa_data:fsa_components(r(Module),Len,Starts,Finals0,_Trans0,[],Fa0),
    LenF is Len + 1,
    lists:append(Finals0,[LenF],FinalsF),
    fsa_data:fsa_components(r(Module,fsa_weights),LenF,Starts,FinalsF,Trans,[],Fa),
    fsa_data:index_finals(Fa0,FinalsIx),
    fsa_data:index_transitions(Fa0,TransIx),
    treat_states(0,Len,LenF,FinalsIx,TransIx,Trans,[],Module).

treat_states(Len0,Len,LenF,FinalsIx,TransIx,Trans0,Trans,Module) :-
    (   Len0 =:= Len
    ->  Trans0=Trans
    ;   treat_state(Len0,LenF,FinalsIx,TransIx,Trans0,Trans1,Module),
	Len1 is Len0 + 1,
	treat_states(Len1,Len,LenF,FinalsIx,TransIx,Trans1,Trans,Module)
    ).

treat_state(Len0,LenF,FinalsIx,TransIx,Trans0,Trans,Module) :-
    findall(Sym/Target,fsa_data:indexed_transition(Len0,TransIx,Sym,Target),Targets),
    count_targets(Targets,Counts,0,Count0,TransIx,FinalsIx,Module),
    fsa_data:indexed_final(Len0,FinalsIx,Bool),
    (   Bool == no
    ->  Trans1=Trans0,
	Count = Count0
    ;   Count is Count0 + 1,
	treat_finals(Count,Len0,LenF,Trans0,Trans1)
    ),
    treat_transitions(Targets,Counts,Count,Len0,Trans1,Trans).

treat_transitions([],[],_,_,Trans,Trans).
treat_transitions([Sym/Target|T],[Count|Counts],Sum,Len0,[trans(Len0,Sym/Weight,Target)|Trans0],Trans) :-
    Weight is Count/Sum,
    treat_transitions(T,Counts,Sum,Len0,Trans0,Trans).
    

treat_finals(Count,Len0,LenF,[trans(Len0,[]/Prob,LenF)|Trans],Trans) :-
    Prob is 1/Count.

count_targets([],[],C,C,_,_,_).
count_targets([Sym/Target|T],[C1C2|Counts],C0,C,TransIx,FinalsIx,Module) :-
    count_target_sym(Sym,C1,Module),
    count_target_state(Target,C2,TransIx,FinalsIx,Module),
    C1C2 is C1 * C2,
    C3 is C0 + C1C2,
    count_targets(T,Counts,C3,C,TransIx,FinalsIx,Module).

count_target_sym(Pred,C,Module) :-
    findall(Sym,fsa_pred_module:evaluate_predicate(Module,Pred,Sym),List),
    length(List,C).

:- dynamic count_target_done/2.

count_target_state(Target,Count,TransIx,FinalsIx,Module) :-
    (   count_target_done(Target,Count0)
    ->  Count0 = Count
    ;   count_target_state0(Target,Count,TransIx,FinalsIx,Module),
	assertz(count_target_done(Target,Count))
    ).

count_target_state0(Len0,Count,TransIx,FinalsIx,Module) :-
    findall(Sym/Target,fsa_data:indexed_transition(Len0,TransIx,Sym,Target),Targets),
    count_targets(Targets,_,0,Count0,TransIx,FinalsIx,Module),
    fsa_data:indexed_final(Len0,FinalsIx,Bool),
    (   Bool == no
    ->  Count = Count0
    ;   Count is Count0 + 1
    ).

